#!/usr/bin/env python
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           Aaron Gibson   0x2C4B239DD876C9F6 <aagbsn@torproject.org>
#           Nick Mathewson 0x21194EBB165733EA <nickm@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2013, The Tor Project, Inc.
#             (c) 2007-2013, all entities within the AUTHORS file
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

from __future__ import print_function

import os
import setuptools
import sys

from glob import glob

# Fix circular dependency with setup.py install
try:
    from babel.messages.frontend import compile_catalog, extract_messages
    from babel.messages.frontend import init_catalog, update_catalog
except ImportError:
    compile_catalog = extract_messages = init_catalog = update_catalog = None

# setup automatic versioning (see top-level versioneer.py file):
import versioneer
versioneer.versionfile_source = 'lib/bridgedb/_version.py'
versioneer.versionfile_build = 'bridgedb/_version.py'

# when creating a release, tags should be prefixed with 'bridgedb-', like so:
#
#     git checkout -b release-6.6.6 develop
#     [do some stuff, merge whatever, test things]
#     git tag -S bridgedb-6.6.6
#     git push tpo-common --tags
#     git checkout master
#     git merge -S --no-ff release-6.6.6
#     git checkout develop
#     git merge -S --no-ff master
#     git branch -d release-6.6.6
#
versioneer.tag_prefix = 'bridgedb-'
# source tarballs should unpack to a directory like 'bridgedb-6.6.6'
versioneer.parentdir_prefix = 'bridgedb-'

pkgpath = os.path.join('lib', 'bridgedb')

# Repo directory that contains translations; this directory should contain
# both uncompiled translations (.po files) as well as compiled ones (.mo
# files). We only want to install the .mo files.
repo_i18n = os.path.join(pkgpath, 'i18n')

# The list of country codes for supported languages will be stored as a list
# variable, ``_supported``, in this file, so that the bridgedb packages
# __init__.py can access it:
repo_langs = os.path.join(pkgpath, '_langs.py')

# The directory containing template files and other resources to serve on the
# web server:
repo_templates = os.path.join(pkgpath, 'templates')

# The directories to install non-sourcecode resources into should always be
# given as relative paths, in order to force distutils to install relative to
# the rest of the codebase.
#
# Directory to installed compiled translations (.mo files) into:
install_i18n = os.path.join('bridgedb', 'i18n')

# Directory to install docs, license, and other text resources into:
install_docs = os.path.join('share', 'doc', 'bridgedb')

def get_cmdclass():
    """Get our cmdclass dictionary for use in setuptool.setup().

    This must be done outside the call to setuptools.setup() because we need
    to add our own classes to the cmdclass dictionary, and then update that
    dictionary with the one returned from versioneer.get_cmdclass().
    """
    cmdclass = {'test': runTests,
                'compile_catalog': compile_catalog,
                'extract_messages': extract_messages,
                'init_catalog': init_catalog,
                'update_catalog': update_catalog}
    cmdclass.update(versioneer.get_cmdclass())
    return cmdclass

def get_requirements():
    """Extract the list of requirements from our requirements.txt.

    :rtype: 2-tuple
    :returns: Two lists, the first is a list of requirements in the form of
        pkgname==version. The second is a list of URIs or VCS checkout strings
        which specify the dependency links for obtaining a copy of the
        requirement.
    """
    requirements_file = os.path.join(os.getcwd(), 'requirements.txt')
    requirements = []
    links=[]
    try:
        with open(requirements_file) as reqfile:
            for line in reqfile.readlines():
                line = line.strip()
                if line.startswith('#'):
                    continue
                elif line.startswith(
                        ('https://', 'git://', 'hg://', 'svn://')):
                    links.append(line)
                else:
                    requirements.append(line)

    except (IOError, OSError) as error:
        print(error)

    return requirements, links

def get_supported_langs():
    """Get the paths for all compiled translation files.

    The two-letter country code of each language which is going to be
    installed will be added to a list, and this list will be written to
    :attr:`repo_langs`, so that lib/bridgedb/__init__.py can store a
    package-level attribute ``bridgedb.__langs__``, which will be a list of
    any languages which were installed.

    Then, the paths of the compiled translations files are added to
    :ivar:`data_files`. These should be included in the ``data_files``
    parameter in :func:`~setuptools.setup` in order for setuptools to be able
    to tell the underlying distutils ``install_data`` command to include these
    files.

    See http://docs.python.org/2/distutils/setupscript.html#installing-additional-files
    for more information.

    :ivar list supported: A list of two-letter country codes, one for each
        language we currently provide translations support for.
    :ivar list lang_dirs: The directories (relative or absolute) to install
        the compiled translation file to.
    :ivar list lang_files: The paths to compiled translations files, relative
        to this setup.py script.
    :rtype: list
    :returns: Two lists, ``lang_dirs`` and ``lang_files``.
    """
    supported = []
    lang_dirs = []
    lang_files = []

    for lang in os.listdir(repo_i18n):
        if lang.endswith('templates'):
            continue
        supported.append(lang)
        lang_dirs.append(os.path.join(install_i18n, lang))
        lang_files.append(os.path.join(repo_i18n, lang,
                                       'LC_MESSAGES', 'bridgedb.mo'))
    supported.sort()

    # Write our list of supported languages to 'lib/bridgedb/_langs.py':
    new_langs_lines = []
    with open(repo_langs, 'r') as langsfile:
        for line in langsfile.readlines():
            if line.startswith('supported'):
                # Change the 'supported' list() into a set():
                line = "supported = set(%s)\n" % supported
            new_langs_lines.append(line)
    with open(repo_langs, 'w') as newlangsfile:
        for line in new_langs_lines:
            newlangsfile.write(line)

    return lang_dirs, lang_files

def get_template_files():
    """Return the paths to any web resource files to include in the package.

    :rtype: list
    :returns: Any files in :attr:`repo_templates` which match one of the glob
        patterns in :ivar:`include_patterns`.
    """
    include_patterns = ['*.html',
                        '*.txt',
                        '*.asc',
                        'assets/*.png',
                        'assets/*.svg',
                        'assets/css/*.css',
                        'assets/font/*.woff',
                        'assets/font/*.ttf',
                        'assets/font/*.svg',
                        'assets/font/*.eot']
    template_files = []

    for include_pattern in include_patterns:
        pattern = os.path.join(repo_templates, include_pattern)
        matches = glob(pattern)
        template_files.extend(matches)

    return template_files

def get_data_files(filesonly=False):
    """Return any hard-coded data_files which should be distributed.

    This is necessary so that both the distutils-derived :class:`installData`
    class and the setuptools ``data_files`` parameter include the same files.
    Call this function with ``filesonly=True`` to get a list of files suitable
    for giving to the ``package_data`` parameter in ``setuptools.setup()``.
    Or, call it with ``filesonly=False`` (the default) to get a list which is
    suitable for using as ``distutils.command.install_data.data_files``.

    :param bool filesonly: If true, only return the locations of the files to
        install, not the directories to install them into.
    :rtype: list
    :returns: If ``filesonly``, returns a list of file paths. Otherwise,
        returns a list of 2-tuples containing: one, the directory to install
        to, and two, the files to install to that directory.
    """
    data_files = []
    doc_files = ['README', 'TODO', 'LICENSE', 'requirements.txt']
    lang_dirs, lang_files = get_supported_langs()
    template_files = get_template_files()

    if filesonly:
        data_files.extend(doc_files)
        for lst in lang_files, template_files:
            for filename in lst:
                if filename.startswith(pkgpath):
                    # The +1 gets rid of the '/' at the beginning:
                    filename = filename[len(pkgpath) + 1:]
                    data_files.append(filename)
    else:
        data_files.append((install_docs, doc_files))
        for ldir, lfile in zip(lang_dirs, lang_files):
            data_files.append((ldir, [lfile,]))

    #[sys.stdout.write("Added data_file '%s'\n" % x) for x in data_files]

    return data_files


class runTests(setuptools.Command):
    # Based on setup.py from mixminion, which is based on setup.py
    # from Zooko's pyutil package, which is in turn based on
    # http://mail.python.org/pipermail/distutils-sig/2002-January/002714.html
    description = "Run unit tests"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        build = self.get_finalized_command('build')
        self.build_purelib = build.build_purelib
        self.build_platlib = build.build_platlib

    def run(self):
        self.run_command('build')
        old_path = sys.path[:]
        sys.path[0:0] = [self.build_purelib, self.build_platlib]
        try:
            testmod = __import__("bridgedb.Tests", globals(), "", [])
            testmod.Tests.main()
        finally:
            sys.path = old_path


# If there is an environment variable BRIDGEDB_INSTALL_DEPENDENCIES=0, it will
# disable checking for, fetching, and installing BridgeDB's dependencies with
# easy_install.
#
# Setting BRIDGEDB_INSTALL_DEPENDENCIES=0 is *highly* recommended, because
# easy_install is a security nightmare.  Automatically installing dependencies
# is enabled by default, however, because this is how all Python packages are
# supposed to work.
if bool(int(os.environ.get("BRIDGEDB_INSTALL_DEPENDENCIES", 1))):
    requires, deplinks = get_requirements()
else:
    requires, deplinks = [], []


setuptools.setup(
    name='bridgedb',
    version=versioneer.get_version(),
    description='Backend systems for distribution of Tor bridge relays',
    author='Nick Mathewson',
    author_email='nickm at torproject dot org',
    maintainer='Isis Agora Lovecruft',
    maintainer_email='isis@torproject.org 0xA3ADB67A2CDB8B35',
    url='https://www.torproject.org',
    download_url='https://gitweb.torproject.org/bridgedb.git',
    package_dir={'': 'lib'},
    packages=['bridgedb',
              'bridgedb.email',
              'bridgedb.parse',
              'bridgedb.test'],
    scripts=['scripts/bridgedb',
             'scripts/get-tor-exits'],
    extras_require={'test': ["sure==1.2.2",
                             "coverage==3.7.1",
                             "leekspin==1.1.4"]},
    zip_safe=False,
    cmdclass=get_cmdclass(),
    include_package_data=True,
    install_requires=requires,
    dependency_links=deplinks,
    package_data={'bridgedb': get_data_files(filesonly=True)},
    exclude_package_data={'bridgedb': ['*.po', '*.pot']},
    message_extractors={pkgpath: [
        ('**.py', 'python', None),
        ('templates/**.html', 'mako', None),
        ('public/**', 'ignore', None)]},
)
# XXX I think we don't need the 'public/**' babel.messages.frontend.method_map
# anymore... 2013-10-15 --isis



#!/usr/bin/env python
"""Unittests for the :mod:`bridgedb.Storage` module."""

from twisted.python import log
from twisted.trial import unittest
import bridgedb.Storage as Storage
from twisted.internet import reactor
from twisted.internet.threads import deferToThread
import os
import threading
from time import sleep

class DatabaseTest(unittest.TestCase):
    def setUp(self):
        self.dbfname = 'test-bridgedb.sqlite'
        Storage.setDBFilename(self.dbfname)

    def tearDown(self):
        if os.path.isfile(self.dbfname):
            os.unlink(self.dbfname)
        Storage.clearGlobalDB()

    def _runAndDie(self, timeout, func):
        with func():
            sleep(timeout)

    def _cb_assertTrue(self, result):
        self.assertTrue(result)

    def _cb_assertFalse(self, result):
        self.assertFalse(result)

    def _eb_Failure(self, failure):
        self.fail(failure)

    def test_getDB_FalseWhenLocked(self):
        Storage._LOCK = threading.Lock()
        Storage._LOCK.acquire()
        self.assertFalse(Storage._LOCK.acquire(False))

    def test_getDB_AcquireLock(self):
        Storage.initializeDBLock()
        with Storage.getDB() as db:
            self.assertIsInstance(db, Storage.Database)
            self.assertTrue(Storage.dbIsLocked())
            self.assertEqual(db, Storage._OPENED_DB)

    def test_getDB_ConcurrencyLock(self):
        timeout = 1
        d1 = deferToThread(self._runAndDie, timeout, Storage.getDB)
        d1.addCallback(self._cb_assertFalse)
        d1.addErrback(self._eb_Failure)
        d2 = deferToThread(Storage.getDB, False)
        d2.addCallback(self._cb_assertFalse)
        d2.addErrback(self._eb_Failure)
        d2.addCallback(self._cb_assertTrue, Storage.getDB(False))



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013, Isis Lovecruft
#             (c) 2007-2013, The Tor Project, Inc.
#             (c) 2007-2013, all entities within the AUTHORS file
# :license: 3-Clause BSD, see LICENSE for licensing information

"""Unittests for the :mod:`bridgedb.persistent` module.

These tests are meant to ensure that the :mod:`bridgedb.persistent` module is
functioning as expected.
"""

from __future__ import print_function

import os.path

from copy import deepcopy
from io   import StringIO

from bridgedb import persistent
from bridgedb.parse.options import MainOptions
from twisted.python import log
from twisted.trial import unittest

import sure
from sure import this
from sure import the
from sure import expect


TEST_CONFIG_FILE = StringIO(unicode("""\
BRIDGE_FILES = ['bridge-descriptors', 'bridge-descriptors.new']
LOGFILE = 'bridgedb.log'"""))


class StateTest(unittest.TestCase):
    """Tests for :class:`bridgedb.persistent.State`."""

    timeout = 15

    def setUp(self):
        configuration = {}
        TEST_CONFIG_FILE.seek(0)
        compiled = compile(TEST_CONFIG_FILE.read(), '<string>', 'exec')
        exec compiled in configuration
        config = persistent.Conf(**configuration)

        fakeArgs = ['-c', os.path.join(os.getcwdu(), '..', 'bridgedb.conf')]
        options = MainOptions()
        options.parseOptions(fakeArgs)

        self.options = options
        self.config  = config
        self.state   = persistent.State(**config.__dict__)
        self.state.options = options
        self.state.config = config

    def test_configCreation(self):
        this(self.config).should.be.ok
        this(self.config).should.be.a(persistent.Conf)

    def test_optionsCreation(self):
        this(self.options).should.be.ok
        this(self.options).should.be.a(dict)

    def test_stateCreation(self):
        this(self.state).should.be.ok

        this(self.state).should.have.property('config').being.ok
        this(self.state).should.have.property('config').being.equal(self.config)

        this(self.state.options).should.be.ok
        this(self.state.options).should.equal(self.options)

    def test_docstring_persistent(self):
        persistent.should.have.property('__doc__').being.a(str)

    def test_docstring_persistentState(self):
        the(self.state).should.have.property('__doc__').being.a(str)

    def test_state_init(self):
        this(self.state).should.have.property('config')
        this(self.state).should.have.property('proxyList')
        this(self.state).should.have.property('statefile')

    def test_persistent_getState(self):
        persistent.should.have.property('_getState').being(callable)
        this(persistent._getState()).should.be.a(persistent.State)

    def test_getStateFor(self):
        jellyState = self.state.getStateFor(self)
        expect(jellyState).to.be.a(dict)
        expect(jellyState.keys()).to.contain('LOGFILE')

    def test_STATEFILE(self):
        this(self.state).should.have.property('statefile')
        the(self.state.statefile).should.be.a(str)

    def test_existsSave(self):
        this(self.state).should.have.property('save').being(callable)

    def test_existsLoad(self):
        persistent.should.have.property('load').being(callable)

    def test_persistent_state(self):
        the(persistent._state).should.be.a(persistent.State)

    def test_before_useChangedSettings_state(self):
        this(self.state).shouldnt.have.property('FOO')
        this(self.state).shouldnt.have.property('BAR')
        this(self.state).should.have.property('LOGFILE').being.a(str)
        this(self.state).should.have.property(
            'BRIDGE_FILES').being.a(list)

    def test_before_useChangedSettings_config(self):
        this(self.config).shouldnt.have.property('FOO')
        this(self.config).shouldnt.have.property('BAR')
        this(self.config).should.have.property('LOGFILE').being.a(str)
        this(self.config).should.have.property(
            'BRIDGE_FILES').being.a(list)

    def test_before_useChangedSettings_stateConfig(self):
        this(self.state.config).shouldnt.have.property('FOO')
        this(self.state.config).shouldnt.have.property('BAR')
        this(self.state.config).should.have.property('LOGFILE').being.a(str)
        this(self.state.config).should.have.property(
            'BRIDGE_FILES').being.a(list)

    def test_useChangedSettings(self):
        # This deepcopying must be done to avoid changing the State object
        # which is used for the rest of the tests.

        thatConfig = deepcopy(self.config)
        thatState  = deepcopy(self.state)

        setattr(thatConfig, 'FOO', 'fuuuuu')
        setattr(thatConfig, 'BAR', 'all of the things')
        setattr(thatConfig, 'LOGFILE', 42)

        this(thatConfig).should.have.property('FOO').being.a(basestring)
        this(thatConfig).should.have.property('BAR').being.a(basestring)
        this(thatConfig).should.have.property('LOGFILE').being.an(int)
        this(thatConfig).should.have.property('BRIDGE_FILES').being.a(list)

        the(thatConfig.FOO).must.equal('fuuuuu')
        the(thatConfig.BAR).must.equal('all of the things')
        the(thatConfig.LOGFILE).must.equal(42)

        the(thatState).should.have.property('useChangedSettings')
        the(thatState.useChangedSettings).should.be(callable)
        thatState.useChangedSettings(thatConfig)

        the(thatState.FOO).should.equal('fuuuuu')
        the(thatState).should.have.property('FOO').being.a(basestring)
        the(thatState).should.have.property('BAR').being.a(basestring)
        the(thatState).should.have.property('LOGFILE').being.an(int)
        the(thatState.FOO).must.equal(thatConfig.FOO)
        the(thatState.BAR).must.equal(thatConfig.BAR)
        the(thatState.LOGFILE).must.equal(thatConfig.LOGFILE)

        this(thatState.config).should.have.property('FOO')
        this(thatState.config).should.have.property('BAR')
        this(thatState.config).should.have.property('LOGFILE').being.an(int)
        this(thatState.config).should.have.property(
            'BRIDGE_FILES').being.a(list)



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_HTTPServer -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: please see included AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2013-2015, Isis Lovecruft
# :license: see LICENSE for licensing information

"""
This module implements the web (http, https) interfaces to the bridge database.
"""

import base64
import gettext
import logging
import random
import re
import textwrap
import time
import os

from functools import partial

from ipaddr import IPv4Address
from ipaddr import IPv6Address
from ipaddr import IPAddress

import mako.exceptions
from mako.template import Template
from mako.lookup import TemplateLookup

from twisted.internet import reactor
from twisted.internet.error import CannotListenError
from twisted.python import filepath
from twisted.web import resource
from twisted.web import server
from twisted.web import static
from twisted.web.util import redirectTo

import bridgedb.Dist
import bridgedb.geo

from bridgedb import captcha
from bridgedb import crypto
from bridgedb import strings
from bridgedb import translations
from bridgedb import txrecaptcha
from bridgedb.Filters import filterBridgesByIP4
from bridgedb.Filters import filterBridgesByIP6
from bridgedb.Filters import filterBridgesByTransport
from bridgedb.Filters import filterBridgesByNotBlockedIn
from bridgedb.parse import headers
from bridgedb.parse.addr import isIPAddress
from bridgedb.qrcodes import generateQR
from bridgedb.safelog import logSafely


TEMPLATE_DIR = os.path.join(os.path.dirname(__file__), 'templates')
rtl_langs = ('ar', 'he', 'fa', 'gu_IN', 'ku')

# Setting `filesystem_checks` to False is recommended for production servers,
# due to potential speed increases. This means that the atimes of the Mako
# template files aren't rechecked every time the template is requested
# (otherwise, if they are checked, and the atime is newer, the template is
# recompiled). `collection_size` sets the number of compiled templates which
# are cached before the least recently used ones are removed. See:
# http://docs.makotemplates.org/en/latest/usage.html#using-templatelookup
lookup = TemplateLookup(directories=[TEMPLATE_DIR],
                        output_encoding='utf-8',
                        filesystem_checks=False,
                        collection_size=500)
logging.debug("Set template root to %s" % TEMPLATE_DIR)

def replaceErrorPage(error, template_name=None):
    """Create a general error page for displaying in place of tracebacks.

    Log the error to BridgeDB's logger, and then display a very plain "Sorry!
    Something went wrong!" page to the client.

    :type error: :exc:`Exception`
    :param error: Any exeption which has occurred while attempting to retrieve
                  a template, render a page, or retrieve a resource.
    :param str template_name: A string describing which template/page/resource
                              was being used when the exception occurred,
                              i.e. ``'index.html'``.
    :returns: A string containing HTML to serve to the client (rather than
              serving a traceback).
    """
    logging.error("Error while attempting to render %s: %s"
                  % (template_name or 'template',
                     mako.exceptions.text_error_template().render()))

    # TRANSLATORS: Please DO NOT translate the following words and/or phrases in
    # any string (regardless of capitalization and/or punctuation):
    #
    # "BridgeDB"
    # "pluggable transport"
    # "pluggable transports"
    # "obfs2"
    # "obfs3"
    # "scramblesuit"
    # "fteproxy"
    # "Tor"
    # "Tor Browser"
    #
    errmsg = _("Sorry! Something went wrong with your request.")
    rendered = """<html>
                    <head>
                      <link href="/assets/bootstrap.min.css" rel="stylesheet">
                      <link href="/assets/custom.css" rel="stylesheet">
                    </head>
                    <body>
                      <p>{0}</p>
                    </body>
                  </html>""".format(errmsg)

    return rendered


class CaptchaProtectedResource(resource.Resource):
    """A general resource protected by some form of CAPTCHA."""

    isLeaf = True

    def __init__(self, publicKey=None, secretKey=None,
                 useForwardedHeader=False, protectedResource=None):
        resource.Resource.__init__(self)
        self.publicKey = publicKey
        self.secretKey = secretKey
        self.useForwardedHeader = useForwardedHeader
        self.resource = protectedResource

    def getClientIP(self, request):
        """Get the client's IP address from the :header:`X-Forwarded-For`
        header, or from the :api:`request <twisted.web.server.Request>`.

        :rtype: None or str
        :returns: The client's IP address, if it was obtainable.
        """
        ip = None
        if self.useForwardedHeader:
            h = request.getHeader("X-Forwarded-For")
            if h:
                ip = h.split(",")[-1].strip()
                if not isIPAddress(ip):
                    logging.warn("Got weird X-Forwarded-For value %r" % h)
                    ip = None
        else:
            ip = request.getClientIP()
        return ip

    def getCaptchaImage(self, request=None):
        """Get a CAPTCHA image.

        :returns: A 2-tuple of ``(image, challenge)``, where ``image`` is a
                  binary, JPEG-encoded image, and ``challenge`` is a unique
                  string. If unable to retrieve a CAPTCHA, returns a tuple
                  containing two empty strings.
        """
        return ('', '')

    def extractClientSolution(self, request):
        """Extract the client's CAPTCHA solution from a POST request.

        This is used after receiving a POST request from a client (which
        should contain their solution to the CAPTCHA), to extract the solution
        and challenge strings.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for 'bridges.html'.
        :returns: A redirect for a request for a new CAPTCHA if there was a
            problem. Otherwise, returns a 2-tuple of strings, the first is the
            client's CAPTCHA solution from the text input area, and the second
            is the challenge string.
        """
        try:
            challenge = request.args['captcha_challenge_field'][0]
            response = request.args['captcha_response_field'][0]
        except Exception:  # pragma: no cover
            return redirectTo(request.URLPath(), request)
        return (challenge, response)

    def checkSolution(self, request):
        """Override this method to check a client's CAPTCHA solution.

        :rtype: bool
        :returns: ``True`` if the client correctly solved the CAPTCHA;
            ``False`` otherwise.
        """
        return False

    def render_GET(self, request):
        """Retrieve a ReCaptcha from the API server and serve it to the client.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for a page which should be
            protected by a CAPTCHA.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
            for the client to solve.
        """
        rtl = False
        image, challenge = self.getCaptchaImage(request)

        try:
            langs = translations.getLocaleFromHTTPRequest(request)
            rtl = translations.usingRTLLang(langs)
            # TODO: this does not work for versions of IE < 8.0
            imgstr = 'data:image/jpeg;base64,%s' % base64.b64encode(image)
            template = lookup.get_template('captcha.html')
            rendered = template.render(strings,
                                       rtl=rtl,
                                       lang=langs[0],
                                       imgstr=imgstr,
                                       challenge_field=challenge)
        except Exception as err:
            rendered = replaceErrorPage(err, 'captcha.html')

        request.setHeader("Content-Type", "text/html; charset=utf-8")
        return rendered

    def render_POST(self, request):
        """Process a client's CAPTCHA solution.

        If the client's CAPTCHA solution is valid (according to
        :meth:`checkSolution`), process and serve their original
        request. Otherwise, redirect them back to a new CAPTCHA page.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
            for the client to solve.
        """
        request.setHeader("Content-Type", "text/html; charset=utf-8")

        if self.checkSolution(request) is True:
            try:
                rendered = self.resource.render(request)
            except Exception as err:
                rendered = replaceErrorPage(err)
            return rendered

        logging.debug("Client failed a CAPTCHA; returning redirect to %s"
                      % request.uri)
        return redirectTo(request.uri, request)


class GimpCaptchaProtectedResource(CaptchaProtectedResource):
    """A web resource which uses a local cache of CAPTCHAs, generated with
    gimp-captcha_, to protect another resource.

    .. _gimp-captcha: https://github.com/isislovecruft/gimp-captcha
    """

    def __init__(self, hmacKey=None, captchaDir='', **kwargs):
        """Protect a resource via this one, using a local CAPTCHA cache.

        :param str secretkey: A PKCS#1 OAEP-padded, private RSA key, used for
            verifying the client's solution to the CAPTCHA. See
            :func:`bridgedb.crypto.getRSAKey` and the
            ``GIMP_CAPTCHA_RSA_KEYFILE`` config setting.
        :param str publickey: A PKCS#1 OAEP-padded, public RSA key, used for
            creating the ``captcha_challenge_field`` string to give to a
            client.
        :param bytes hmacKey: The master HMAC key, used for validating CAPTCHA
            challenge strings in :meth:`captcha.GimpCaptcha.check`. The file
            where this key is stored can be set via the
            ``GIMP_CAPTCHA_HMAC_KEYFILE`` option in the config file.
        :param str captchaDir: The directory where the cached CAPTCHA images
            are stored. See the ``GIMP_CAPTCHA_DIR`` config setting.
        :param bool useForwardedHeader: If ``True``, obtain the client's IP
            address from the ``X-Forwarded-For`` HTTP header.
        :type protectedResource: :api:`twisted.web.resource.Resource`
        :param protectedResource: The resource to serve if the client
            successfully passes the CAPTCHA challenge.
        """
        CaptchaProtectedResource.__init__(self, **kwargs)
        self.hmacKey = hmacKey
        self.captchaDir = captchaDir

    def checkSolution(self, request):
        """Process a solved CAPTCHA via :meth:`bridgedb.captcha.GimpCaptcha.check`.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :rtupe: bool
        :returns: True, if the CAPTCHA solution was valid; False otherwise.
        """
        valid = False
        challenge, solution = self.extractClientSolution(request)
        clientIP = self.getClientIP(request)
        clientHMACKey = crypto.getHMAC(self.hmacKey, clientIP)

        try:
            valid = captcha.GimpCaptcha.check(challenge, solution,
                                              self.secretKey, clientHMACKey)
        except captcha.CaptchaExpired as error:
            logging.warn(error)
            valid = False

        logging.debug("%sorrect captcha from %r: %r."
                      % ("C" if valid else "Inc", clientIP, solution))
        return valid

    def getCaptchaImage(self, request):
        """Get a random CAPTCHA image from our **captchaDir**.

        Creates a :class:`~bridgedb.captcha.GimpCaptcha`, and calls its
        :meth:`~bridgedb.captcha.GimpCaptcha.get` method to return a random
        CAPTCHA and challenge string.

        :type request: :api:`twisted.web.http.Request`
        :param request: A client's initial request for some other resource
            which is protected by this one (i.e. protected by a CAPTCHA).
        :returns: A 2-tuple of ``(image, challenge)``, where::
            - ``image`` is a string holding a binary, JPEG-encoded image.
            - ``challenge`` is a unique string associated with the request.
        """
        # Create a new HMAC key, specific to requests from this client:
        clientIP = self.getClientIP(request)
        clientHMACKey = crypto.getHMAC(self.hmacKey, clientIP)
        capt = captcha.GimpCaptcha(self.publicKey, self.secretKey,
                                   clientHMACKey, self.captchaDir)
        try:
            capt.get()
        except captcha.GimpCaptchaError as error:
            logging.error(error)
        except Exception as error:  # pragma: no cover
            logging.error("Unhandled error while retrieving Gimp captcha!")
            logging.exception(error)

        return (capt.image, capt.challenge)

    def render_GET(self, request):
        """Get a random CAPTCHA from our local cache directory and serve it to
        the client.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for a page which should be
            protected by a CAPTCHA.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
           for the client to solve.
        """
        return CaptchaProtectedResource.render_GET(self, request)

    def render_POST(self, request):
        """Process a client's CAPTCHA solution.

        If the client's CAPTCHA solution is valid (according to
        :meth:`checkSolution`), process and serve their original
        request. Otherwise, redirect them back to a new CAPTCHA page.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
            for the client to solve.
        """
        return CaptchaProtectedResource.render_POST(self, request)


class ReCaptchaProtectedResource(CaptchaProtectedResource):
    """A web resource which uses the reCaptcha_ service.

    .. _reCaptcha: http://www.google.com/recaptcha
    """

    def __init__(self, remoteIP=None, **kwargs):
        CaptchaProtectedResource.__init__(self, **kwargs)
        self.remoteIP = remoteIP

    def _renderDeferred(self, checkedRequest):
        """Render this resource asynchronously.

        :type checkedRequest: tuple
        :param checkedRequest: A tuple of ``(bool, request)``, as returned
            from :meth:`checkSolution`.
        """
        try:
            valid, request = checkedRequest
        except Exception as err:
            logging.error("Error in _renderDeferred(): %s" % err)
            return

        logging.debug("Attemping to render %svalid request %r"
                      % ('' if valid else 'in', request))
        if valid is True:
            try:
                rendered = self.resource.render(request)
            except Exception as err:  # pragma: no cover
                rendered = replaceErrorPage(err)
        else:
            logging.info("Client failed a CAPTCHA; redirecting to %s"
                         % request.uri)
            rendered = redirectTo(request.uri, request)

        try:
            request.write(rendered)
            request.finish()
        except Exception as err:  # pragma: no cover
            logging.exception(err)

        return request

    def getCaptchaImage(self, request):
        """Get a CAPTCHA image from the remote reCaptcha server.

        :type request: :api:`twisted.web.http.Request`
        :param request: A client's initial request for some other resource
            which is protected by this one (i.e. protected by a CAPTCHA).
        :returns: A 2-tuple of ``(image, challenge)``, where::
            - ``image`` is a string holding a binary, JPEG-encoded image.
            - ``challenge`` is a unique string associated with the request.
        """
        capt = captcha.ReCaptcha(self.publicKey, self.secretKey)

        try:
            capt.get()
        except Exception as error:
            logging.fatal("Connection to Recaptcha server failed: %s" % error)

        if capt.image is None:
            logging.warn("No CAPTCHA image received from ReCaptcha server!")

        return (capt.image, capt.challenge)

    def getRemoteIP(self):
        """Mask the client's real IP address with a faked one.

        The fake client IP address is sent to the reCaptcha server, and it is
        either the public IP address of bridges.torproject.org (if the config
        option ``RECAPTCHA_REMOTE_IP`` is configured), or a random IP.

        :rtype: str
        :returns: A fake IP address to report to the reCaptcha API server.
        """
        if self.remoteIP:
            remoteIP = self.remoteIP
        else:
            # generate a random IP for the captcha submission
            remoteIP = IPv4Address(random.randint(0, 2**32-1)).compressed

        return remoteIP

    def checkSolution(self, request):
        """Process a solved CAPTCHA by sending it to the ReCaptcha server.

        The client's IP address is not sent to the ReCaptcha server; instead,
        a completely random IP is generated and sent instead.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :rtupe: :api:`twisted.internet.defer.Deferred`
        :returns: A deferred which will callback with a tuple in the following
            form:
                (:type:`bool`, :api:`twisted.web.server.Request`)
            If the CAPTCHA solution was valid, a tuple will contain::
                (True, Request)
            Otherwise, it will contain::
                (False, Request)
        """
        challenge, response = self.extractClientSolution(request)
        clientIP = self.getClientIP(request)
        remoteIP = self.getRemoteIP()

        logging.debug("Captcha from %r. Parameters: %r"
                      % (clientIP, request.args))

        def checkResponse(solution, request):
            """Check the :class:`txrecaptcha.RecaptchaResponse`.

            :type solution: :class:`txrecaptcha.RecaptchaResponse`.
            :param solution: The client's CAPTCHA solution, after it has been
                submitted to the reCaptcha API server.
            """
            # This valid CAPTCHA result from this function cannot be reliably
            # unittested, because it's callbacked to from the deferred
            # returned by ``txrecaptcha.submit``, the latter of which would
            # require networking (as well as automated CAPTCHA
            # breaking). Hence, the 'no cover' pragma.
            if solution.is_valid:  # pragma: no cover
                logging.info("Valid CAPTCHA solution from %r." % clientIP)
                return (True, request)
            else:
                logging.info("Invalid CAPTCHA solution from %r: %r"
                             % (clientIP, solution.error_code))
                return (False, request)

        d = txrecaptcha.submit(challenge, response, self.secretKey,
                               remoteIP).addCallback(checkResponse, request)
        return d

    def render_GET(self, request):
        """Retrieve a ReCaptcha from the API server and serve it to the client.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object for 'bridges.html'.
        :rtype: str
        :returns: A rendered HTML page containing a ReCaptcha challenge image
            for the client to solve.
        """
        return CaptchaProtectedResource.render_GET(self, request)

    def render_POST(self, request):
        """Process a client's CAPTCHA solution.

        If the client's CAPTCHA solution is valid (according to
        :meth:`checkSolution`), process and serve their original
        request. Otherwise, redirect them back to a new CAPTCHA page.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object, including POST arguments which
            should include two key/value pairs: one key being
            ``'captcha_challenge_field'``, and the other,
            ``'captcha_response_field'``. These POST arguments should be
            obtained from :meth:`render_GET`.
        :returns: :api:`twisted.web.server.NOT_DONE_YET`, in order to handle
            the ``Deferred`` returned from :meth:`checkSolution`. Eventually,
            when the ``Deferred`` request is done being processed,
            :meth:`_renderDeferred` will handle rendering and displaying the
            HTML to the client.
        """
        d = self.checkSolution(request)
        d.addCallback(self._renderDeferred)
        return server.NOT_DONE_YET


class WebResourceOptions(resource.Resource):
    """A resource with additional options which a client may use to specify the
    which bridge types should be returned by :class:`WebResourceBridges`.
    """

    isLeaf = True

    def __init__(self):
        """Create a new WebResource for the Options page"""
        gettext.install("bridgedb", unicode=True)
        resource.Resource.__init__(self)

    def render_GET(self, request):
        rtl = False
        try:
            langs = translations.getLocaleFromHTTPRequest(request)
            rtl = translations.usingRTLLang(langs)
            template = lookup.get_template('options.html')
            rendered = template.render(strings, rtl=rtl, lang=langs[0])
        except Exception as err:  # pragma: no cover
            rendered = replaceErrorPage(err)
        request.setHeader("Content-Type", "text/html; charset=utf-8")
        return rendered

    render_POST = render_GET


class WebResourceHowto(resource.Resource):
    """A resource which explains how to use bridges."""

    isLeaf = True

    def __init__(self):
        """Create a new WebResource for the Options page"""
        gettext.install("bridgedb", unicode=True)
        resource.Resource.__init__(self)

    def render_GET(self, request):
        rtl = False
        try:
            langs = translations.getLocaleFromHTTPRequest(request)
            rtl = translations.usingRTLLang(langs)
            template = lookup.get_template('howto.html')
            rendered = template.render(strings, rtl=rtl, lang=langs[0])
        except Exception as err:  # pragma: no cover
            rendered = replaceErrorPage(err)
        request.setHeader("Content-Type", "text/html; charset=utf-8")
        return rendered

    render_POST = render_GET


class WebResourceBridges(resource.Resource):
    """This resource displays bridge lines in response to a request."""

    isLeaf = True

    def __init__(self, distributor, schedule, N=1, useForwardedHeader=False,
                 includeFingerprints=True):
        """Create a new resource for displaying bridges to a client.

        :type distributor: :class:`IPBasedDistributor`
        :param distributor: The mechanism to retrieve bridges for this
            distributor.
        :type schedule: :class:`~bridgedb.schedule.ScheduledInterval`
        :param schedule: The time period used to tweak the bridge selection
            procedure.
        :param int N: The number of bridges to hand out per query.
        :param bool useForwardedHeader: Whether or not we should use the the
            X-Forwarded-For header instead of the source IP address.
        :param bool includeFingerprints: Do we include the bridge's
            fingerprint in the response?
        """
        gettext.install("bridgedb", unicode=True)
        resource.Resource.__init__(self)
        self.distributor = distributor
        self.schedule = schedule
        self.nBridgesToGive = N
        self.useForwardedHeader = useForwardedHeader
        self.includeFingerprints = includeFingerprints

    def render(self, request):
        """Render a response for a client HTTP request.

        Presently, this method merely wraps :meth:`getBridgeRequestAnswer` to
        catch any unhandled exceptions which occur (otherwise the server will
        display the traceback to the client). If an unhandled exception *does*
        occur, the client will be served the default "No bridges currently
        available" HTML response page.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object containing the HTTP method, full
            URI, and any URL/POST arguments and headers present.
        :rtype: str
        :returns: A plaintext or HTML response to serve.
        """
        try:
            response = self.getBridgeRequestAnswer(request)
        except Exception as err:
            logging.exception(err)
            response = self.renderAnswer(request)

        return response

    def getBridgeRequestAnswer(self, request):
        """Respond to a client HTTP request for bridges.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object containing the HTTP method, full
            URI, and any URL/POST arguments and headers present.
        :rtype: str
        :returns: A plaintext or HTML response to serve.
        """
        # XXX why are we getting the interval if our distributor might be
        # using bridgedb.schedule.Unscheduled?
        interval = self.schedule.intervalStart(time.time())
        bridges = ( )
        ip = None
        countryCode = None

        # XXX this code is duplicated in CaptchaProtectedResource
        if self.useForwardedHeader:
            h = request.getHeader("X-Forwarded-For")
            if h:
                ip = h.split(",")[-1].strip()
                if not isIPAddress(ip):
                    logging.warn("Got weird forwarded-for value %r",h)
                    ip = None
        else:
            ip = request.getClientIP()

        # Look up the country code of the input IP
        if isIPAddress(ip):
            countryCode = bridgedb.geo.getCountryCode(IPAddress(ip))
        else:
            logging.warn("Invalid IP detected; skipping country lookup.")
            countryCode = None

        # XXX separate function again
        format = request.args.get("format", None)
        if format and len(format): format = format[0] # choose the first arg

        # do want any options?
        transport = ipv6 = unblocked = False

        ipv6 = request.args.get("ipv6", False)
        if ipv6: ipv6 = True # if anything after ?ipv6=

        # XXX oh dear hell. why not check for the '?transport=' arg before
        # regex'ing? And why not compile the regex once, somewhere outside
        # this function and class?
        try:
            # validate method name
            transport = re.match('[_a-zA-Z][_a-zA-Z0-9]*',
                    request.args.get("transport")[0]).group()
        except (TypeError, IndexError, AttributeError):
            transport = None

        try:
            unblocked = re.match('[a-zA-Z]{2,4}',
                    request.args.get("unblocked")[0]).group()
        except (TypeError, IndexError, AttributeError):
            unblocked = False

        logging.info("Replying to web request from %s. Parameters were %r"
                     % (ip, request.args))

        rules = []
        bridgeLines = None

        if ip:
            if ipv6:
                rules.append(filterBridgesByIP6)
                addressClass = IPv6Address
            else:
                rules.append(filterBridgesByIP4)
                addressClass = IPv4Address

            if transport:
                #XXX: A cleaner solution would differentiate between
                # addresses by protocol rather than have separate lists
                # Tor to be a transport, and selecting between them.
                rules = [filterBridgesByTransport(transport, addressClass)]

            if unblocked:
                rules.append(filterBridgesByNotBlockedIn(unblocked,
                    addressClass, transport))

            bridges = self.distributor.getBridgesForIP(ip, interval,
                                                       self.nBridgesToGive,
                                                       countryCode,
                                                       bridgeFilterRules=rules)
            bridgeLines = "".join("%s\n" % b.getConfigLine(
                includeFingerprint=self.includeFingerprints,
                addressClass=addressClass,
                transport=transport,
                request=bridgedb.Dist.uniformMap(ip)
                ) for b in bridges)

        answer = self.renderAnswer(request, bridgeLines, format)
        return answer

    def renderAnswer(self, request, bridgeLines=None, format=None):
        """Generate a response for a client which includes **bridges**.

        The generated response can be plaintext or HTML.

        :type request: :api:`twisted.web.http.Request`
        :param request: A ``Request`` object containing the HTTP method, full
            URI, and any URL/POST arguments and headers present.
        :type bridgeLines: list or None
        :param bridgeLines: A list of strings used to configure a Tor client
            to use a bridge.
        :type format: str or None
        :param format: If ``'plain'``, return a plaintext response. Otherwise,
            use the :file:`bridgedb/templates/bridges.html` template to render
            an HTML response page which includes the **bridges**.
        :rtype: str
        :returns: A plaintext or HTML response to serve.
        """
        rtl = False

        if format == 'plain':
            request.setHeader("Content-Type", "text/plain")
            rendered = bridgeLines
        else:
            request.setHeader("Content-Type", "text/html; charset=utf-8")
            qrcode = None
            qrjpeg = generateQR(bridgeLines)

            if qrjpeg:
                qrcode = 'data:image/jpeg;base64,%s' % base64.b64encode(qrjpeg)
            try:
                langs = translations.getLocaleFromHTTPRequest(request)
                rtl = translations.usingRTLLang(langs)
                template = lookup.get_template('bridges.html')
                rendered = template.render(strings,
                                           rtl=rtl,
                                           lang=langs[0],
                                           answer=bridgeLines,
                                           qrcode=qrcode)
            except Exception as err:
                rendered = replaceErrorPage(err)

        return rendered


class WebRoot(resource.Resource):
    """The parent resource of all other documents hosted by the webserver."""

    isLeaf = True

    def render_GET(self, request):
        """Handles requests for the webserver root document.

        For example, this function handles requests for
        https://bridges.torproject.org/.

        :type request: :api:`twisted.web.server.Request`
        :param request: An incoming request.
        """
        rtl = False
        try:
            langs = translations.getLocaleFromHTTPRequest(request)
            rtl = translations.usingRTLLang(langs)
            template = lookup.get_template('index.html')
            rendered = template.render(strings,
                                       rtl=rtl,
                                       lang=langs[0])
        except Exception as err:
            rendered = replaceErrorPage(err)

        return rendered


def addWebServer(cfg, dist, sched):
    """Set up a web server for HTTP(S)-based bridge distribution.

    :type cfg: :class:`bridgedb.persistent.Conf`
    :param cfg: A configuration object from
         :mod:`bridgedb.Main`. Currently, we use these options::
             HTTP_UNENCRYPTED_PORT
             HTTP_UNENCRYPTED_BIND_IP
             HTTP_USE_IP_FROM_FORWARDED_HEADER
             HTTPS_N_BRIDGES_PER_ANSWER
             HTTPS_INCLUDE_FINGERPRINTS
             HTTPS_KEY_FILE
             HTTPS_CERT_FILE
             HTTPS_PORT
             HTTPS_BIND_IP
             HTTPS_USE_IP_FROM_FORWARDED_HEADER
             RECAPTCHA_ENABLED
             RECAPTCHA_PUB_KEY
             RECAPTCHA_SEC_KEY
             RECAPTCHA_REMOTEIP
             GIMP_CAPTCHA_ENABLED
             GIMP_CAPTCHA_DIR
             GIMP_CAPTCHA_HMAC_KEYFILE
             GIMP_CAPTCHA_RSA_KEYFILE
    :type dist: :class:`bridgedb.Dist.IPBasedDistributor`
    :param dist: A bridge distributor.
    :type sched: :class:`bridgedb.schedule.ScheduledInterval`
    :param sched: The scheduled interval at which bridge selection, which
        are ultimately displayed on the :class:`WebResourceBridges` page, will
        be shifted.
    :raises SystemExit: if the servers cannot be started.
    :rtype: :api:`twisted.web.server.Site`
    :returns: A webserver.
    """
    captcha = None
    fwdHeaders = cfg.HTTP_USE_IP_FROM_FORWARDED_HEADER
    numBridges = cfg.HTTPS_N_BRIDGES_PER_ANSWER
    fprInclude = cfg.HTTPS_INCLUDE_FINGERPRINTS

    logging.info("Starting web servers...")

    httpdist = resource.Resource()
    httpdist.putChild('', WebRoot())
    httpdist.putChild('robots.txt',
                      static.File(os.path.join(TEMPLATE_DIR, 'robots.txt')))
    httpdist.putChild('keys',
                      static.File(os.path.join(TEMPLATE_DIR, 'bridgedb.asc')))
    httpdist.putChild('assets',
                      static.File(os.path.join(TEMPLATE_DIR, 'assets/')))
    httpdist.putChild('options', WebResourceOptions())
    httpdist.putChild('howto', WebResourceHowto())

    if cfg.RECAPTCHA_ENABLED:
        publicKey = cfg.RECAPTCHA_PUB_KEY
        secretKey = cfg.RECAPTCHA_SEC_KEY
        captcha = partial(ReCaptchaProtectedResource,
                          remoteIP=cfg.RECAPTCHA_REMOTEIP)
    elif cfg.GIMP_CAPTCHA_ENABLED:
        # Get the master HMAC secret key for CAPTCHA challenges, and then
        # create a new HMAC key from it for use on the server.
        captchaKey = crypto.getKey(cfg.GIMP_CAPTCHA_HMAC_KEYFILE)
        hmacKey = crypto.getHMAC(captchaKey, "Captcha-Key")
        # Load or create our encryption keys:
        secretKey, publicKey = crypto.getRSAKey(cfg.GIMP_CAPTCHA_RSA_KEYFILE)
        captcha = partial(GimpCaptchaProtectedResource,
                          hmacKey=hmacKey,
                          captchaDir=cfg.GIMP_CAPTCHA_DIR)

    bridges = WebResourceBridges(dist, sched, numBridges,
                                 fwdHeaders, includeFingerprints=fprInclude)
    if captcha:
        # Protect the 'bridges' page with a CAPTCHA, if configured to do so:
        protected = captcha(publicKey=publicKey,
                            secretKey=secretKey,
                            useForwardedHeader=fwdHeaders,
                            protectedResource=bridges)
        httpdist.putChild('bridges', protected)
        logging.info("Protecting resources with %s." % captcha.func.__name__)
    else:
        httpdist.putChild('bridges', bridges)

    site = server.Site(httpdist)
    site.displayTracebacks = False

    if cfg.HTTP_UNENCRYPTED_PORT:
        ip = cfg.HTTP_UNENCRYPTED_BIND_IP or ""
        port = cfg.HTTP_UNENCRYPTED_PORT or 80
        try:
            reactor.listenTCP(port, site, interface=ip)
        except CannotListenError as error:
            raise SystemExit(error)
        logging.info("Started HTTP server on %s:%d" % (str(ip), int(port)))

    if cfg.HTTPS_PORT:
        ip = cfg.HTTPS_BIND_IP or ""
        port = cfg.HTTPS_PORT or 443
        try:
            from twisted.internet.ssl import DefaultOpenSSLContextFactory
            factory = DefaultOpenSSLContextFactory(cfg.HTTPS_KEY_FILE,
                                                   cfg.HTTPS_CERT_FILE)
            reactor.listenSSL(port, site, factory, interface=ip)
        except CannotListenError as error:
            raise SystemExit(error)
        logging.info("Started HTTPS server on %s:%d" % (str(ip), int(port)))

    return site



# -*- coding: utf-8 -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2014-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""Tests for :mod:`bridgedb.qrcodes`."""


from twisted.trial import unittest

from bridgedb import qrcodes


class GenerateQRTests(unittest.TestCase):
    """Unittests for :func:`bridgedb.qrcodes.generateQR`."""

    def setUp(self):
        self.qrcodeModule = qrcodes.qrcode
        bridgelines = [
            "obfs4 63.125.48.205:26573 441a151632806a3cc42adfecc2e6e823299db7b2 iat-mode=1 public-key=8d27ba37c5d810106b55f3fd6cdb35842007e88754184bfc0e6035f9bcede633 node-id=42d2a6ad49f93ab4b987b1a9e738425aacb8d2af",
            "obfs4 103.111.131.45:43288 cb1362f8eaf5d3c2b6fad5da300f66c4197f23e5 iat-mode=0 public-key=c75cb66ae28d8ebc6eded002c28a8ba0d06d3a78c6b5cbf9b2ade051f0775ac4 node-id=4cd66dfabbd964f8c6c4414b07cdb45dae692e19",
            "obfs4 17.194.28.21:5530 86ef8ab343e76bcd3c57ee32febe4482c98141c7 iat-mode=0 public-key=36ebe205bcdfc499a25e6923f4450fa8d48196ceb4fa0ce077d9d8ec4a36926d node-id=6b6277afcb65d33525545904e95c2fa240632660",
        ]
        self.bridgelines = '\n'.join(bridgelines)

    def tearDown(self):
        """Replace the qrcode module to its original form."""
        qrcodes.qrcode = self.qrcodeModule

    def test_generateQR(self):
        """Calling generateQR() should generate an image."""
        self.assertTrue(qrcodes.generateQR(self.bridgelines))

    def test_generateQR_bad_bridgelines(self):
        """Calling generateQR() with a bad type for the bridgelines should
        return None.
        """
        self.assertIsNone(qrcodes.generateQR(list()))

    def test_generateQR_no_bridgelines(self):
        """Calling generateQR() without bridgelines should return None."""
        self.assertIsNone(qrcodes.generateQR(""))

    def test_generateQR_no_qrcode_module(self):
        """Calling generateQR() without the qrcode module installed should
        return None.
        """
        qrcodes.qrcode = None
        self.assertIsNone(qrcodes.generateQR(self.bridgelines))

    def test_generateQR_bridgeSchema(self):
        """Calling generateQR() with bridgeSchema=True should prepend
        ``'bridge://`` to each of the QR encoded bridge lines.
        """
        # If we were to install the python-qrtools Debian package, we'd be
        # able to decode the resulting QRCode to check that it contains the
        # 'bridge://' prefix for each bridge line… but that would add another
        # Debian dependency just to unittest 5 lines of code.
        #
        # Instead:
        self.assertTrue(qrcodes.generateQR(self.bridgelines, bridgeSchema=True))

    def test_generateQR_save_nonexistent_format(self):
        """Calling generateQR() with imageFormat=u'FOOBAR' should return None.
        """
        self.assertIsNone(qrcodes.generateQR(self.bridgelines, imageFormat=u'FOOBAR'))



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2013, The Tor Project, Inc.
#             (c) 2007-2013, all entities within the AUTHORS file
# :license: 3-clause BSD, see included LICENSE for information

"""Parsers for BridgeDB commandline options.

.. py:module:: bridgedb.parse.options
   :synopsis: Parsers for BridgeDB commandline options.


bridgedb.parse.options
======================
::

  bridgedb.parse.options
   |__ setConfig()
   |__ getConfig() - Set/Get the config file path.
   |__ setRundir()
   |__ getRundir() - Set/Get the runtime directory.
   |__ parseOptions() - Create the main options parser for BridgeDB.
   |
   \_ BaseOptions - Base options, included in all other options menus.
       ||
       |\__ findRundirAndConfigFile() - Find the absolute path of the config
       |                                file and runtime directory, or find
       |                                suitable defaults.
       |
       |__ SIGHUPOptions - Menu to explain SIGHUP signal handling and usage.
       |__ SIGUSR1Options - Menu to explain SIGUSR1 handling and usage.
       |
       |__ MockOptions - Suboptions for creating fake bridge descriptors for
       |                 testing purposes.
       |__ TestOptions - Suboptions for running twisted.trial based unittests.
       \__ MainOptions - Main commandline options parser for BridgeDB.
..
"""

from __future__ import print_function
from __future__ import unicode_literals

import sys
import textwrap
import traceback
import os

from twisted.python import usage

from bridgedb import __version__


#: In :meth:`BaseOptions.findRundirAndConfig`, this is set to the the
#: absolute path of the ``opts['rundir']`` setting, if given, otherwise it
#: defaults to the current directory.
_rundir = None

#: In :meth:`BaseOptions.findRundirAndConfig`, if ``opts['config']`` is
#: given, this is set to the the absolute path of the ``opts['config']``
#: settting relative to the ``rundir``, otherwise it defaults to
#: 'bridgedb.conf' in the current directory.
_config = None


def setConfig(path):
    """Set the absolute path to the config file.

    See :meth:`BaseOptions.postOptions`.

    :param string path: The path to set.
    """
    global _config
    _config = path

def getConfig():
    """Get the absolute path to the config file.

    :rtype: string
    :returns: The path to the config file.
    """
    return _config

def setRundir(path):
    """Set the absolute path to the runtime directory.

    See :meth:`BaseOptions.postOptions`.

    :param string path: The path to set.
    """
    global _rundir
    _rundir = path

def getRundir():
    """Get the absolute path to the runtime directory.

    :rtype: string
    :returns: The path to the config file.
    """
    return _rundir

def parseOptions():
    """Create the main options parser and its subcommand parsers.

    Any :exc:`~twisted.python.usage.UsageErrors` which are raised due to
    invalid options are ignored; their error message is printed and then we
    exit the program.

    :rtype: :class:`MainOptions`
    :returns: The main options parsing class, with any commandline arguments
        already parsed.
    """
    options = MainOptions()

    try:
        options.parseOptions()
    except usage.UsageError as uerr:
        print(uerr.message)
        print(options.getUsage())
        sys.exit(1)
    except Exception as error:  # pragma: no cover
        exc, value, tb = sys.exc_info()
        print("Unhandled Error: %s" % error.message)
        print(traceback.format_exc(tb))

    return options


class BaseOptions(usage.Options):
    """Base options included in all main and sub options menus."""

    longdesc = textwrap.dedent("""BridgeDB is a proxy distribution system for
    private relays acting as bridges into the Tor network. See `bridgedb
    <command> --help` for addition help.""")

    optParameters = [
        ['config', 'c', None,
         'Configuration file [default: <rundir>/bridgedb.conf]'],
        ['rundir', 'r', None,
         """Change to this directory before running. [default: `os.getcwd()']

         All other paths, if not absolute, should be relative to this path.
         This includes the config file and any further files specified within
         the config file.
         """]]

    def __init__(self):
        """Create an options parser. All flags, parameters, and attributes of
        this base options parser are inherited by all child classes.
        """
        super(BaseOptions, self).__init__()
        self['version'] = self.opt_version
        self['verbosity'] = 30

    def opt_quiet(self):
        """Decrease verbosity"""
        # We use '10' because then it corresponds to the log levels
        self['verbosity'] -= 10

    def opt_verbose(self):
        """Increase verbosity"""
        self['verbosity'] += 10

    opt_q = opt_quiet
    opt_v = opt_verbose

    def opt_version(self):
        """Display BridgeDB's version and exit."""
        print("%s-%s" % (__package__, __version__))
        sys.exit(0)

    @staticmethod
    def findRundirAndConfigFile(rundir=None, config=None):
        """Find the absolute path of the config file and runtime directory, or
        find suitable defaults.

        Attempts to set the absolute path of the runtime directory. If the
        config path is relative, its absolute path is set relative to the
        runtime directory path (unless it starts with '.' or '..', then it is
        interpreted relative to the current working directory). If the path to
        the config file is absolute, it is left alone.

        :type rundir: string or None
        :param rundir: The user-supplied path to the runtime directory, from
            the commandline options (i.e.
            ``options = BaseOptions().parseOptions(); options['rundir'];``).
        :type config: string or None
        :param config: The user-supplied path to the config file, from the
            commandline options (i.e.
            ``options = BaseOptions().parseOptions(); options['config'];``).
        :raises: :api:`twisted.python.usage.UsageError` if either the runtime
            directory or the config file cannot be found.
        """
        gRundir = getRundir()
        gConfig = getConfig()

        if gRundir is None:
            if rundir is not None:
                gRundir = os.path.abspath(os.path.expanduser(rundir))
            else:
                gRundir = os.getcwdu()
        setRundir(gRundir)

        if not os.path.isdir(gRundir):  # pragma: no cover
            raise usage.UsageError(
                "Could not change to runtime directory: `%s'" % gRundir)

        if gConfig is None:
            if config is None:
                config = 'bridgedb.conf'
            gConfig = config

            if not os.path.isabs(gConfig):
                # startswith('.') will handle other relative paths, i.e. '..'
                if gConfig.startswith('.'):  # pragma: no cover
                    gConfig = os.path.abspath(os.path.expanduser(gConfig))
                else:
                    gConfig = os.path.join(gRundir, gConfig)
        setConfig(gConfig)

        gConfig = getConfig()
        if not os.path.isfile(gConfig):  # pragma: no cover
            raise usage.UsageError(
                "Specified config file `%s' doesn't exist!" % gConfig)

    def postOptions(self):
        """Automatically called by :meth:`parseOptions`.

        Determines appropriate values for the 'config' and 'rundir' settings.
        """
        super(BaseOptions, self).postOptions()
        self.findRundirAndConfigFile(self['rundir'], self['config'])

        gConfig = getConfig()
        gRundir = getRundir()

        if (self['rundir'] is None) and (gRundir is not None):
            self['rundir'] = gRundir

        if (self['config'] is None) and (gConfig is not None):
            self['config'] = gConfig

        if self['verbosity'] <= 10:
            print("%s.postOptions():" % self.__class__)
            print("  gCONFIG=%s" % gConfig)
            print("  self['config']=%s" % self['config'])
            print("  gRUNDIR=%s" % gRundir)
            print("  self['rundir']=%s" % self['rundir'])


class TestOptions(BaseOptions):
    """Suboptions for running twisted.trial and unittest based tests."""

    longdesc = textwrap.dedent("""BridgeDB testing commands.
    See the `bridgedb mock` command for generating testing environments.""")

    optFlags = [['coverage', 'c', 'Generate coverage statistics']]
    optParameters = [
        ['file', 'f', None, 'Run tests in specific file(s) (trial only)'],
        ['unittests', 'u', False, 'Run unittests in bridgedb.Tests'],
        ['trial', 't', True, 'Run twisted.trial tests in bridgedb.test']]

    completionData = usage.Completions(
        mutuallyExclusive=[('unittests', 'coverage'),
                           ('unittests', 'file')],
        optActions={'file': usage.CompleteFiles('lib/bridgedb/test/test_*.py',
                                                repeat=True,
                                                descr="test filename")},
        extraActions=[
            usage.Completer(descr="extra arguments to pass to trial")])

    def parseArgs(self, *args):
        """Parse any additional arguments after the options and flags."""
        self['test_args'] = args


class MockOptions(BaseOptions):
    """Suboptions for creating necessary conditions for testing purposes."""

    optParameters = [
        ['descriptors', 'n', 1000,
         '''Generate <n> mock bridge descriptor sets
          (types: netstatus, extrainfo, server)''']]


class SIGHUPOptions(BaseOptions):
    """Options menu to explain usage and handling of SIGHUP signals."""

    longdesc = """If you send a SIGHUP to a running BridgeDB process, the
    servers will parse and reload all bridge descriptor files into the
    databases.

    Note that this command WILL NOT handle sending the signal for you; see
    signal(7) and kill(1) for additional help."""


class SIGUSR1Options(BaseOptions):
    """Options menu to explain usage and handling of SIGUSR1 signals."""

    longdesc = """If you send a SIGUSR1 to a running BridgeDB process, the
    servers will dump all bridge assignments by distributor from the
    databases to files.

    Note that this command WILL NOT handle sending the signal for you; see
    signal(7) and kill(1) for additional help."""


class MainOptions(BaseOptions):
    """Main commandline options parser for BridgeDB."""

    optFlags = [
        ['dump-bridges', 'd', 'Dump bridges by hashring assignment into files'],
        ['reload', 'R', 'Reload bridge descriptors into running servers']]
    subCommands = [
        ['test', None, TestOptions, "Run twisted.trial tests or unittests"],
        ['mock', None, MockOptions, "Generate a testing environment"],
        ['SIGHUP', None, SIGHUPOptions,
         "Reload bridge descriptors into running servers"],
        ['SIGUSR1', None, SIGUSR1Options,
         "Dump bridges by hashring assignment into files"]]



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013-2015 Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
# :license: 3-clause BSD, see included LICENSE for information

"""Parsers for bridge nicknames.

.. py:module:: bridgedb.parse.nickname
   :synopsis: Parsers for Tor bridge nicknames.

bridgedb.parse.nicknames
========================
::

  nicknames
   |_ isValidRouterNickname - Determine if a nickname is according to spec
..
"""

import logging
import string


class InvalidRouterNickname(ValueError):
    """Router nickname doesn't follow tor-spec."""


def isValidRouterNickname(nickname):
    """Determine if a router's given nickname meets the specification.

    :param string nickname: An OR's nickname.
    :rtype: bool
    :returns: ``True`` if the nickname is valid, ``False`` otherwise.
    """
    ALPHANUMERIC = string.letters + string.digits

    try:
        if not (1 <= len(nickname) <= 19):
            raise InvalidRouterNickname(
                "Nicknames must be between 1 and 19 characters: %r" % nickname)
        for letter in nickname:
            if not letter in ALPHANUMERIC:
                raise InvalidRouterNickname(
                    "Nicknames must only use [A-Za-z0-9]: %r" % nickname)
    except InvalidRouterNickname as error:
        logging.error(str(error))
    except TypeError:  # The nickname was probably still set to ``None``
        pass
    else:
        return True

    return False



# BridgeDB by Nick Mathewson.
# Copyright (c) 2007-2009, The Tor Project, Inc.
# See LICENSE for licensing information

import calendar
import os
import logging
import binascii
import sqlite3
import time
import hashlib
from contextlib import contextmanager
from contextlib import GeneratorContextManager
from functools import wraps
from ipaddr import IPAddress, IPv6Address, IPv4Address
import sys

import bridgedb.Stability as Stability
from bridgedb.Stability import BridgeHistory
import threading

toHex = binascii.b2a_hex
fromHex = binascii.a2b_hex
HEX_ID_LEN = 40

def _escapeValue(v):
    return "'%s'" % v.replace("'", "''")

def timeToStr(t):
    return time.strftime("%Y-%m-%d %H:%M", time.gmtime(t))
def strToTime(t):
    return calendar.timegm(time.strptime(t, "%Y-%m-%d %H:%M"))

#  The old DB system was just a key->value mapping DB, with special key
#  prefixes to indicate which database they fell into.
#
#     sp|<ID> -- given to bridgesplitter; maps bridgeID to ring name.
#     em|<emailaddr> -- given to emailbaseddistributor; maps email address
#            to concatenated ID.
#     fs|<ID> -- Given to BridgeTracker, maps to time when a router was
#            first seen (YYYY-MM-DD HH:MM)
#     ls|<ID> -- given to bridgetracker, maps to time when a router was
#            last seen (YYYY-MM-DD HH:MM)
#
# We no longer want to use em| at all, since we're not doing that kind
# of persistence any more.

# Here is the SQL schema.
SCHEMA2_SCRIPT = """
 CREATE TABLE Config (
     key PRIMARY KEY NOT NULL,
     value
 );

 CREATE TABLE Bridges (
     id INTEGER PRIMARY KEY NOT NULL,
     hex_key,
     address,
     or_port,
     distributor,
     first_seen,
     last_seen
 );

 CREATE UNIQUE INDEX BridgesKeyIndex ON Bridges ( hex_key );

 CREATE TABLE EmailedBridges (
     email PRIMARY KEY NOT NULL,
     when_mailed
 );

 CREATE INDEX EmailedBridgesWhenMailed on EmailedBridges ( email );

 CREATE TABLE BlockedBridges (
     id INTEGER PRIMARY KEY NOT NULL,
     hex_key,
     blocking_country
 );

 CREATE INDEX BlockedBridgesBlockingCountry on BlockedBridges(hex_key);

 CREATE TABLE WarnedEmails (
     email PRIMARY KEY NOT NULL,
     when_warned
 );

 CREATE INDEX WarnedEmailsWasWarned on WarnedEmails ( email );

 INSERT INTO Config VALUES ( 'schema-version', 2 ); 
"""

SCHEMA_2TO3_SCRIPT = """
 CREATE TABLE BridgeHistory (
     fingerprint PRIMARY KEY NOT NULL,
     address,
     port INT,
     weightedUptime LONG,
     weightedTime LONG,
     weightedRunLength LONG,
     totalRunWeights DOUBLE,
     lastSeenWithDifferentAddressAndPort LONG,
     lastSeenWithThisAddressAndPort LONG,
     lastDiscountedHistoryValues LONG,
     lastUpdatedWeightedTime LONG
 );

 CREATE INDEX BridgeHistoryIndex on BridgeHistory ( fingerprint );

 INSERT OR REPLACE INTO Config VALUES ( 'schema-version', 3 ); 
 """
SCHEMA3_SCRIPT = SCHEMA2_SCRIPT + SCHEMA_2TO3_SCRIPT

class BridgeData:
    """Value class carrying bridge information:
       hex_key      - The unique hex key of the given bridge
       address      - Bridge IP address
       or_port      - Bridge TCP port
       distributor  - The distributor (or pseudo-distributor) through which 
                      this bridge is being announced
       first_seen   - When did we first see this bridge online?
       last_seen    - When was the last time we saw this bridge online?
    """
    def __init__(self, hex_key, address, or_port, distributor="unallocated", 
                 first_seen="", last_seen=""):
        self.hex_key = hex_key
        self.address = address
        self.or_port = or_port
        self.distributor = distributor
        self.first_seen = first_seen
        self.last_seen = last_seen

class Database(object):
    def __init__(self, sqlite_fname, db_fname=None):
        if db_fname is None:
            self._conn = openDatabase(sqlite_fname)
        else:
            self._conn = openOrConvertDatabase(sqlite_fname, db_fname)
        self._cur = self._conn.cursor()
        self.sqlite_fname = sqlite_fname

    def commit(self):
        self._conn.commit()

    def rollback(self):
        self._conn.rollback()

    def close(self):
        #print "Closing DB"
        self._cur.close()
        self._conn.close()

    def insertBridgeAndGetRing(self, bridge, setRing, seenAt, validRings,
                               defaultPool="unallocated"):
        '''Updates info about bridge, setting ring to setRing if none was set.
           Also sets distributor to `defaultPool' if the bridge was found in
           the database, but its distributor isn't valid anymore.

           Returns the name of the distributor the bridge is assigned to.
        '''
        cur = self._cur

        t = timeToStr(seenAt)
        h = bridge.fingerprint
        assert len(h) == HEX_ID_LEN

        cur.execute("SELECT id, distributor "
                    "FROM Bridges WHERE hex_key = ?", (h,))
        v = cur.fetchone()
        if v is not None:
            i, ring = v
            # Check if this is currently a valid ring name. If not, move back
            # into default pool.
            if ring not in validRings:
                ring = defaultPool
            # Update last_seen, address, port and (possibly) distributor.
            cur.execute("UPDATE Bridges SET address = ?, or_port = ?, "
                        "distributor = ?, last_seen = ? WHERE id = ?",
                        (str(bridge.address), bridge.orPort, ring,
                         timeToStr(seenAt), i))
            return ring
        else:
            # Insert it.
            cur.execute("INSERT INTO Bridges (hex_key, address, or_port, "
                        "distributor, first_seen, last_seen) "
                        "VALUES (?, ?, ?, ?, ?, ?)",
                        (h, str(bridge.address), bridge.orPort, setRing, t, t))
            return setRing

    def cleanEmailedBridges(self, expireBefore):
        cur = self._cur
        t = timeToStr(expireBefore)

        cur.execute("DELETE FROM EmailedBridges WHERE when_mailed < ?", (t,))

    def getEmailTime(self, addr):
        addr = hashlib.sha1(addr).hexdigest()
        cur = self._cur
        cur.execute("SELECT when_mailed FROM EmailedBridges WHERE "
                    "email = ?", (addr,))
        v = cur.fetchone()
        if v is None:
            return None
        return strToTime(v[0])

    def setEmailTime(self, addr, whenMailed):
        addr = hashlib.sha1(addr).hexdigest()
        cur = self._cur
        t = timeToStr(whenMailed)
        cur.execute("INSERT OR REPLACE INTO EmailedBridges "
                    "(email,when_mailed) VALUES (?,?)", (addr, t))

    def getAllBridges(self):
        """Return a list of BridgeData value classes of all bridges in the
           database
        """
        retBridges = []
        cur = self._cur
        cur.execute("SELECT hex_key, address, or_port, distributor, "
                    "first_seen, last_seen  FROM Bridges")
        for b in cur.fetchall():
            bridge = BridgeData(b[0], b[1], b[2], b[3], b[4], b[5])
            retBridges.append(bridge)

        return retBridges

    def getBridgesForDistributor(self, distributor):
        """Return a list of BridgeData value classes of all bridges in the
           database that are allocated to distributor 'distributor'
        """
        retBridges = []
        cur = self._cur
        cur.execute("SELECT hex_key, address, or_port, distributor, "
                    "first_seen, last_seen FROM Bridges WHERE "
                    "distributor = ?", (distributor, ))
        for b in cur.fetchall():
            bridge = BridgeData(b[0], b[1], b[2], b[3], b[4], b[5])
            retBridges.append(bridge)

        return retBridges

    def updateDistributorForHexKey(self, distributor, hex_key):
        cur = self._cur
        cur.execute("UPDATE Bridges SET distributor = ? WHERE hex_key = ?",
                    (distributor, hex_key))

    def addBridgeBlock(self, fingerprint, countryCode):
        cur = self._cur
        cur.execute("INSERT OR REPLACE INTO BlockedBridges "
                    "(hex_key,blocking_country) VALUES (?,?)",
                    (fingerprint, countryCode))

    def delBridgeBlock(self, fingerprint, countryCode):
        cur = self._cur
        cur.execute("DELETE FROM BlockedBridges WHERE hex_key = ? "
                    "AND blocking_country = ?", (fingerprint, countryCode))

    def cleanBridgeBlocks(self):
        cur = self._cur
        cur.execute("DELETE FROM BlockedBridges")

    def getBlockingCountries(self, fingerprint):
        cur = self._cur
        cur.execute("SELECT hex_key, blocking_country FROM BlockedBridges WHERE hex_key = ? ",
                    (fingerprint,))
        v = cur.fetchall()
        if v is None:
            return None

        # return list of country-codes
        return [ str(result[1]) for (result) in v ]

    def getBlockedBridges(self, countryCode):
        cur = self._cur
        cur.execute("SELECT hex_key, blocking_country FROM BlockedBridges WHERE blocking_country = ? ",
                    (countryCode,))
        v = cur.fetchall()
        if v is None:
            return None
        # return list of fingerprints
        return [ str(result[0]) for (result) in v ]

    def isBlocked(self, fingerprint, countryCode):
        cur = self._cur
        cur.execute("SELECT hex_key, blocking_country FROM BlockedBridges WHERE "
                    "hex_key = ? AND blocking_country = ?",
                    (fingerprint, countryCode))
        v = cur.fetchone()
        if v is None:
            return False
        return True 

    def getWarnedEmail(self, addr):
        addr = hashlib.sha1(addr).hexdigest()
        cur = self._cur
        cur.execute("SELECT * FROM WarnedEmails WHERE "
                    " email = ?", (addr,))
        v = cur.fetchone()
        if v is None:
            return False
        return True

    def setWarnedEmail(self, addr, warned=True, whenWarned=time.time()):
        addr = hashlib.sha1(addr).hexdigest()
        t = timeToStr(whenWarned)
        cur = self._cur
        if warned == True:
            cur.execute("INSERT INTO WarnedEmails"
                        "(email,when_warned) VALUES (?,?)", (addr, t,))
        elif warned == False:
            cur.execute("DELETE FROM WarnedEmails WHERE "
                        "email = ?", (addr,))

    def cleanWarnedEmails(self, expireBefore):
        cur = self._cur
        t = timeToStr(expireBefore)

        cur.execute("DELETE FROM WarnedEmails WHERE when_warned < ?", (t,))

    def updateIntoBridgeHistory(self, bh):
        cur = self._cur
        cur.execute("INSERT OR REPLACE INTO BridgeHistory values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (bh.fingerprint, str(bh.ip), bh.port,
                bh.weightedUptime, bh.weightedTime, bh.weightedRunLength,
                bh.totalRunWeights, bh.lastSeenWithDifferentAddressAndPort,
                bh.lastSeenWithThisAddressAndPort, bh.lastDiscountedHistoryValues,
                bh.lastUpdatedWeightedTime))
        return bh

    def delBridgeHistory(self, fp):
        cur = self._cur
        cur.execute("DELETE FROM BridgeHistory WHERE fingerprint = ?", (fp,))

    def getBridgeHistory(self, fp):
        cur = self._cur
        cur.execute("SELECT * FROM BridgeHistory WHERE fingerprint = ?", (fp,))
        h = cur.fetchone()
        if h is None: 
            return
        return BridgeHistory(h[0],IPAddress(h[1]),h[2],h[3],h[4],h[5],h[6],h[7],h[8],h[9],h[10])

    def getAllBridgeHistory(self):
        cur = self._cur
        v = cur.execute("SELECT * FROM BridgeHistory")
        if v is None: return
        fp = {}
        for h in v:
            yield BridgeHistory(h[0],IPAddress(h[1]),h[2],h[3],h[4],h[5],h[6],h[7],h[8],h[9],h[10])

    def getBridgesLastUpdatedBefore(self, statusPublicationMillis):
        cur = self._cur
        v = cur.execute("SELECT * FROM BridgeHistory WHERE lastUpdatedWeightedTime < ?", (statusPublicationMillis,))
        if v is None: return
        for h in v:
            yield BridgeHistory(h[0],IPAddress(h[1]),h[2],h[3],h[4],h[5],h[6],h[7],h[8],h[9],h[10])
def openDatabase(sqlite_file):
    conn = sqlite3.Connection(sqlite_file)
    cur = conn.cursor()
    try:
        try:
            cur.execute("SELECT value FROM Config WHERE key = 'schema-version'")
            val, = cur.fetchone()
            if val == 2:
                logging.info("Adding new table BridgeHistory")
                cur.executescript(SCHEMA_2TO3_SCRIPT)
            elif val != 3:
                logging.warn("Unknown schema version %s in database.", val)
        except sqlite3.OperationalError:
            logging.warn("No Config table found in DB; creating tables")
            cur.executescript(SCHEMA3_SCRIPT)
            conn.commit()
    finally:
        cur.close()
    return conn


def openOrConvertDatabase(sqlite_file, db_file):
    """Open a sqlite database, converting it from a db file if needed."""
    if os.path.exists(sqlite_file):
        return openDatabase(sqlite_file)

    conn = sqlite3.Connection(sqlite_file)
    cur = conn.cursor()
    cur.executescript(SCHEMA3_SCRIPT)
    conn.commit()

    import anydbm

    try:
        db = anydbm.open(db_file, 'r')
    except anydbm.error:
        return conn

    try:
        # We handle all the sp| keys first, since other tables have
        # dependencies on Bridges.
        for k in db.keys():
            v = db[k]
            if k.startswith("sp|"):
                assert len(k) == 23
                cur.execute("INSERT INTO Bridges ( hex_key, distributor ) "
                            "VALUES (?, ?)", (toHex(k[3:]),v))
        # Now we handle the other key types.
        for k in db.keys():
            v = db[k]
            if k.startswith("fs|"):
                assert len(k) == 23
                cur.execute("UPDATE Bridges SET first_seen = ? "
                            "WHERE hex_key = ?", (v, toHex(k[3:])))
            elif k.startswith("ls|"):
                assert len(k) == 23
                cur.execute("UPDATE Bridges SET last_seen = ? "
                            "WHERE hex_key = ?", (v, toHex(k[3:])))
            #elif k.startswith("em|"):
            #    keys = list(toHex(i) for i in
            #        bridgedb.Bridges.chopString(v, bridgedb.Bridges.ID_LEN))
            #    cur.executemany("INSERT INTO EmailedBridges ( email, id ) "
            #                    "SELECT ?, id FROM Bridges WHERE hex_key = ?",
            #                    [(k[3:],i) for i in keys])
            elif k.startswith("sp|") or k.startswith("em|"):
                pass
            else:
                logging.warn("Unrecognized key %r", k)
    except:
        conn.rollback()
        conn.close()
        os.unlink(sqlite_file)
        raise

    conn.commit()
    return conn

class DBGeneratorContextManager(GeneratorContextManager):
    """Helper for @contextmanager decorator.

    Overload __exit__() so we can call the generator many times
    """
    def __exit__(self, type, value, traceback):
        """Handle exiting a with statement block

        Progress generator or throw exception

        Significantly based on contextlib.py

        :throws: `RuntimeError` if the generator doesn't stop after
            exception is thrown
        """
        if type is None:
            try:
                self.gen.next()
            except StopIteration:
                return
            return
        else:
            if value is None:
                # Need to force instantiation so we can reliably
                # tell if we get the same exception back
                value = type()
            try:
                self.gen.throw(type, value, traceback)
                raise RuntimeError("generator didn't stop after throw()")
            except StopIteration, exc:
                # Suppress the exception *unless* it's the same exception that
                # was passed to throw().  This prevents a StopIteration
                # raised inside the "with" statement from being suppressed
                return exc is not value
            except:
                # only re-raise if it's *not* the exception that was
                # passed to throw(), because __exit__() must not raise
                # an exception unless __exit__() itself failed.  But throw()
                # has to raise the exception to signal propagation, so this
                # fixes the impedance mismatch between the throw() protocol
                # and the __exit__() protocol.
                #
                if sys.exc_info()[1] is not value:
                    raise

def contextmanager(func):
    """Decorator to for :func:`Storage.getDB()`

    Define getDB() for use by with statement content manager
    """
    @wraps(func)
    def helper(*args, **kwds):
        return DBGeneratorContextManager(func(*args, **kwds))
    return helper

_DB_FNAME = None
_LOCK = None
_LOCKED = 0
_OPENED_DB = None
_REFCOUNT = 0

def clearGlobalDB():
    """Start from scratch

    This is currently only used in unit tests.
    """
    global _DB_FNAME
    global _LOCK
    global _LOCKED
    global _OPENED_DB

    _DB_FNAME = None
    _LOCK = None
    _LOCKED = 0
    _OPENED_DB = None
    _REFCOUNT = 0

def initializeDBLock():
    """Create the lock

    This must be called before the first database query
    """
    global _LOCK

    if not _LOCK:
        _LOCK = threading.RLock()
    assert _LOCK

def checkAndConvertDB(sqlite_file, db_file):
    openOrConvertDatabase(sqlite_file, db_file).close()

def setDBFilename(sqlite_fname):
    global _DB_FNAME
    _DB_FNAME = sqlite_fname

@contextmanager
def getDB(block=True):
    """Generator: Return a usable database handler

    Always return a :class:`bridgedb.Storage.Database` that is
    usable within the current thread. If a connection already exists
    and it was created by the current thread, then return the
    associated :class:`bridgedb.Storage.Database` instance. Otherwise,
    create a new instance, blocking until the existing connection
    is closed, if applicable.

    Note: This is a blocking call (by default), be careful about
        deadlocks!

    :rtype: :class:`bridgedb.Storage.Database`
    :returns: An instance of :class:`bridgedb.Storage.Database` used to
        query the database
    """
    global _DB_FNAME
    global _LOCK
    global _LOCKED
    global _OPENED_DB
    global _REFCOUNT

    assert _LOCK
    try:
        own_lock = _LOCK.acquire(block)
        if own_lock:
            _LOCKED += 1

            if not _OPENED_DB:
                assert _REFCOUNT == 0
                _OPENED_DB = Database(_DB_FNAME)

            _REFCOUNT += 1
            yield _OPENED_DB
        else:
            yield False
    finally:
        assert own_lock
        try:
            _REFCOUNT -= 1
            if _REFCOUNT == 0:
                _OPENED_DB.close()
                _OPENED_DB = None
        finally:
            _LOCKED -= 1
            _LOCK.release()

def dbIsLocked():
    return _LOCKED != 0
